<?php

namespace Drupal\field_states\Form;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\field\Entity\FieldConfig;

/**
 * Transition form.
 */
class StateTransitionForm extends FormBase implements StateTransitionFormInterface {

  /**
   * The entity.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface
   */
  protected $entity;

  /**
   * The state field name.
   *
   * @var string
   */
  protected $fieldName;

  /**
   * {@inheritdoc}
   */
  public function getEntity() {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function setEntity(ContentEntityInterface $entity) {
    $this->entity = $entity;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldName() {
    return $this->fieldName;
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldName($field_name) {
    $this->fieldName = $field_name;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseFormId() {
    return 'state_machine_transition_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    $entity = $this->getEntity();
    if (!$entity) {
      throw new \RuntimeException('No entity provided to StateTransitionForm.');
    }
    // Example ID: "state_machine_transition_form_commerce_order_state_1".
    $form_id = $this->getBaseFormId();
    $form_id .= '_' . $entity->getEntityTypeId() . '_' . $this->fieldName;
    $form_id .= '_' . $entity->id();

    return $form_id;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\state_machine\Plugin\Field\FieldType\StateItemInterface $state_item */
    $state_item = $this->entity->get($this->fieldName)->first();
    $state = !empty($state_item) ? $state_item->get('value')->getString() : '';
    $form['actions'] = [
      '#type' => 'container',
    ];
    $transitions = [];
    if (!empty($state_item)) {
      $transitions = $state_item->getPossibleTransitions($state);
    }
    else {
      $workflow = $this->getWorkflow();
      if (!empty($workflow['transitions'])) {
        foreach ($workflow['transitions'] as $transition_id => $transition) {
          if (in_array($state, $transition['from'])) {
            $transitions[$transition_id] = $transition;
          }
        }
      }
    }
    foreach ($transitions as $transition_id => $transition) {
      $transition['id'] = $transition_id;
      $allowed = $this->isTransitionAllowed($transition);
      if (!$allowed) {
        continue;
      }
      $form['actions'][$transition_id] = [
        '#type' => 'submit',
        '#value' => $transition['label'],
        '#submit' => ['::submitForm'],
        '#transition' => $transition,
      ];
    }
    if (empty($transitions)) {
      $form['#empty_transition'] = TRUE;
      $form['#state'] = $state;
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function isTransitionAllowed($transition) {
    $current_user = \Drupal::currentUser();
    // Check guard with id plugin guard.
    if (!empty($transition['guard'])) {
      $type = \Drupal::service('plugin.manager.states.guard');
      $plugin = $type->createInstance($transition['guard']);
      if (!empty($plugin) && method_exists($plugin, 'allowed')) {
        return $plugin->allowed($transition, $this->getWorkflow(), $this->getEntity());
      }
    }
    // Check permission.
    if (!empty($transition['permission'])) {
      return $current_user->hasPermission($transition['permission']);
    }
    // Check role.
    $userRoles = $current_user->getRoles();
    if (in_array('administrator', $userRoles)) {
      return TRUE;
    }
    if (!empty($transition['role'])) {
      $workflowRoles = is_array($transition['role']) ? $transition['role'] : [$transition['role']];
      $intersection = array_intersect($workflowRoles, $userRoles);
      return !empty($intersection);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    /** @var \Drupal\state_machine\Plugin\Workflow\WorkflowTransition $transition */
    $transition = $triggering_element['#transition'];
    /** @var \Drupal\state_machine\Plugin\Field\FieldType\StateItemInterface $state_item */
    $state_item = $this->entity->get($this->fieldName)->first();
    // Ensure the transition is still allowed before applying it.
    $transition['state'] = $this->entity->get($this->fieldName)->value;
    $isAllowed = !empty($state_item) && $this->isTransitionAllowed($transition);
    if ($isAllowed) {
      $state_item->applyTransition($transition);
    }
    else {
      $this->entity->get($this->fieldName)
        ->setValue(['value' => $transition['to']]);
      if (!empty($transition['workflow'])) {
        $type = \Drupal::service('plugin.manager.states.workflow');
        $workflows = is_array($transition['workflow']) ? $transition['workflow'] : [$transition['workflow']];
        foreach ($workflows as $workflowId) {
          $plugin = $type->createInstance($workflowId);
          if (!empty($plugin) && method_exists($plugin, 'action')) {
            $plugin->action($transition, $this->getWorkflow(), $this->entity);
          }
        }
      }
    }
    // Save log.
    $field_name = $this->fieldName;
    $properties = [
      'targetEntityType' => $this->getEntity()->getEntityTypeId(),
      'bundle' => $this->getEntity()->bundle(),
    ];
    $currentUser = \Drupal::currentUser();
    $formatter = \Drupal::service('date.formatter');
    $view_displays = \Drupal::entityTypeManager()
      ->getStorage('entity_view_display')
      ->loadByProperties($properties);
    $view_display = current($view_displays);
    $fieldSettings = $view_display->getComponent($field_name)['settings'];
    if (!empty($history = $fieldSettings['history'])) {
      $field_instance = FieldConfig::loadByName($properties['targetEntityType'], $properties['bundle'], $history);
      $fieldType = $field_instance->getFieldStorageDefinition()->getType();

      $timestamp = strtotime('now');
      $log = [
        'date' => $formatter->format($timestamp),
        'from' => $transition["state"],
        'to' => $transition["to"],
        'user' => $currentUser->getDisplayName(),
      ];
      $to = ' → ';
      switch ($fieldType) {
        case 'json':
          $value = $this->entity->get($history)->value;
          $json = json_decode($value ?: '[]');
          $json[] = $log;
          $this->entity->set($history, json_encode($json));
          break;

        case 'string':
          $this->entity->$history->appendItem(
            $log['date'] . ' State: ' . $log['from'] . $to . $log['to'] . ' by: ' . $log['user']
          );
          break;

        case 'double_field':
          $this->entity->$history->appendItem([
            'first' => date('Y-m-d\TH:i:s'),
            'second' => $log['from'] . $to . $log['to'] . ' - ' . $log['user'],
          ]);
          break;

        case 'triples_field':
          $this->entity->$history->appendItem([
            'first' => date('Y-m-d\TH:i:s'),
            'second' => implode($to, [$log['from'], $log['to']]),
            'third' => $log['user'],
          ]);
          break;

      }
    }
    $this->entity->save();
    if ($isAllowed) {
      $state_item->applyAction($transition);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getWorkflow() {
    $fieldName = $this->fieldName;
    $fieldDefinition = $this->getEntity()->getFieldDefinition($fieldName);
    $workflow = trim($fieldDefinition->getSetting('workflows'));
    if (!empty($workflow)) {
      $workflow = Yaml::decode($workflow);
    }
    return $workflow;
  }

}
